use std::collections::BTreeMap;

use snafu::{ResultExt, Snafu};

/// The types and directives in a VijiN project.
#[derive(Debug, PartialEq, Eq)]
pub struct TypeSystem {
    pub types: BTreeMap<String, Type>,
    pub directives: BTreeMap<String, Directive>,
}

/// The definition of a standard directive.
///
/// Note that this type is an input to the parser, not an output.
#[derive(Debug, PartialEq, Eq, Default, Clone)]
pub struct Directive {
    /// The provider's name for the directive.
    name: String,
    /// Description of the directive.
    doc: String,
    /// Fixed prefix term.
    prefix: String,
    /// The type of the prefix parameter, if there is one.
    prefix_param: Option<Param>,
    /// Fixed infix term.
    /// Can only be present if the prefix parameter exists.
    infix: Option<String>,
    /// The type of the infix parameter.
    /// Can only be present if the infix term exists.
    infix_param: Option<Param>,
    /// Fixed suffix term.
    suffix: Option<String>,
    /// The type of the suffix parameter.
    /// Can only be present if the suffix exists.
    suffix_param: Option<Param>,
    /// The type of the elements of the list parameter, if supported.
    list_param: Option<Param>,
    // TODO(zeph): macro expansion?
}

/// Directive parameter.
#[derive(Debug, PartialEq, Eq, Clone)]
pub struct Param {
    /// Name to which the compiler maps the argument.
    name: String,
    /// Description of the parameter's meaning in the context of the directive.
    doc: Option<String>,
    /// Type required for the argument.
    typ: Type,
    /// Nuance parameter for list parameters.
    nuance: Option<Box<Param>>,
}

/// Types expected by directive arguments.
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Type {
    /// Arbitrary text.
    Text,
    /// Speech text, possibly containing emphasis, shout, or whisper markers.
    Speech,
    /// Character in the current scene.
    Character,
    /// Duration.
    Duration,
    /// Asset requiring particular tags.
    Asset { tags: Vec<String> },
    /// Enumeration of literal options.
    Enum(Vec<String>),
}

// Serialized representation of directives.
mod serial {
    use serde::Deserialize;

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Document {
        pub types: Vec<Type>,
        pub directives: Vec<Directive>,
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Type {
        pub name: String,
        pub kind: String,
        pub tags: Option<Vec<String>>,
        #[serde(rename = "enum")]
        pub enm: Option<Vec<String>>,
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Param {
        pub name: String,
        pub doc: Option<String>,
        #[serde(rename = "type")]
        pub typ: String,
        pub nuance: Option<Box<Param>>,
    }

    #[derive(Deserialize)]
    #[serde(rename_all = "snake_case", deny_unknown_fields)]
    pub struct Directive {
        pub name: String,
        pub doc: Option<String>,
        pub prefix: String,
        pub prefix_param: Option<Param>,
        pub infix: Option<String>,
        pub infix_param: Option<Param>,
        pub suffix: Option<String>,
        pub suffix_param: Option<Param>,
        pub list_param: Option<Param>,
    }
}

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("invalid directive definition file: {source}"))]
    Decode { source: serde_json::Error },

    #[snafu(display("type {name} has invalid kind {given}"))]
    InvalidKind { name: String, given: String },

    #[snafu(display("{kind} type {name} is missing its {field}"))]
    IncompleteType {
        name: String,
        kind: String,
        field: &'static str,
    },

    #[snafu(display("no type {name} of {position} parameter for {directive}"))]
    ParamType {
        directive: String,
        position: &'static str,
        name: String,
    },

    #[snafu(display("{directive} has {position} parameter with no {position}"))]
    MissingSyntax {
        directive: String,
        position: &'static str,
    },

    #[snafu(display("{directive} has illegal nuance on {position}"))]
    InvalidNuance {
        directive: String,
        position: &'static str,
    },
}

static DIRECTIVE_PREDECLARED: &str = include_str!("directive-predeclared.json");

impl TypeSystem {
    pub fn new() -> TypeSystem {
        let mut ts = TypeSystem {
            types: BTreeMap::new(),
            directives: BTreeMap::new(),
        };
        ts.parse(DIRECTIVE_PREDECLARED).unwrap();
        ts
    }

    pub fn parse(&mut self, src: &str) -> Result<&mut Self, Error> {
        let s: serial::Document = serde_json::from_str(src).context(DecodeSnafu)?;
        for t in s.types {
            let p = match t.kind.as_str() {
                "text" => Type::Text,
                "speech" => Type::Speech,
                "character" => Type::Character,
                "duration" => Type::Duration,
                "asset" => match t.tags {
                    None => {
                        return Err(Error::IncompleteType {
                            name: t.name,
                            kind: t.kind,
                            field: "tags",
                        });
                    }
                    Some(v) => Type::Asset { tags: v },
                },
                "enum" => match t.enm {
                    None => {
                        return Err(Error::IncompleteType {
                            name: t.name,
                            kind: t.kind,
                            field: "enumerants",
                        });
                    }
                    Some(v) => Type::Enum(v),
                },
                _ => {
                    return Err(Error::InvalidKind {
                        name: t.name,
                        given: t.kind,
                    });
                }
            };
            self.types.insert(t.name, p);
        }
        for d in s.directives {
            let prefix_param = d
                .prefix_param
                .map(|p| self.param(&d.name, "prefix", p))
                .transpose()?;
            let infix_param = d
                .infix_param
                .map(|p| {
                    if d.suffix.is_none() {
                        return Err(Error::MissingSyntax {
                            directive: d.name.clone(),
                            position: "infix",
                        });
                    }
                    self.param(&d.name, "infix", p)
                })
                .transpose()?;
            let suffix_param = d
                .suffix_param
                .map(|p| {
                    if d.suffix.is_none() {
                        return Err(Error::MissingSyntax {
                            directive: d.name.clone(),
                            position: "suffix",
                        });
                    }
                    self.param(&d.name, "suffix", p)
                })
                .transpose()?;
            let list_param = d
                .list_param
                .map(|p| self.param(&d.name, "list", p))
                .transpose()?;
            let r = Directive {
                name: d.name.clone(),
                doc: d.doc.unwrap_or_default(),
                prefix: d.prefix,
                prefix_param,
                infix: d.infix,
                infix_param,
                suffix: d.suffix,
                suffix_param,
                list_param,
            };
            self.directives.insert(d.name, r);
        }
        Ok(self)
    }

    fn param(
        &self,
        directive: &str,
        position: &'static str,
        p: serial::Param,
    ) -> Result<Param, Error> {
        let t = match self.types.get(&p.typ) {
            None => {
                return Err(Error::ParamType {
                    directive: directive.into(),
                    position,
                    name: p.typ,
                })
            }
            Some(t) => t,
        };
        let nuance = if let Some(n) = p.nuance {
            if position != "list" {
                return Err(Error::InvalidNuance {
                    directive: directive.into(),
                    position,
                });
            }
            Some(Box::new(self.param(directive, "list nuance", *n)?))
        } else {
            None
        };
        Ok(Param {
            name: p.name,
            doc: p.doc,
            typ: t.clone(),
            nuance,
        })
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use rstest::rstest;

    macro_rules! vec_into {
        [] => { Vec::new() };
        [$($val:expr),+ $(,)?] =>{
            Vec::from([$($val,)+].map(|x| x.into()))
        };
    }

    #[rstest]
    fn builtin() {
        let mut s = TypeSystem {
            types: BTreeMap::new(),
            directives: BTreeMap::new(),
        };
        s.parse(DIRECTIVE_PREDECLARED).unwrap();
        let expected = TypeSystem{
            types: BTreeMap::from([
                ("text", Type::Text),
                ("speech", Type::Speech),
                ("character", Type::Character),
                ("duration", Type::Duration),
                ("scene", Type::Asset { tags: vec_into!["scene"] }),
                ("background", Type::Asset { tags: vec_into!["background"] }),
                ("stage position", Type::Enum(vec_into!["close", "left third", "right third", "left", "center", "right", "far left", "mid left", "mid right", "far right", "far"])),
            ].map(|(k, v)| (k.into(), v))),
            directives: BTreeMap::from([
                ("page", Directive {
                    name: "page".into(),
                    doc: "Start showing the current static speaker, text, &c.\nNormally not used directly; script speech lines call this directive implicitly.".into(),
                    prefix: "page".into(),
                    ..Default::default()
                }),
                ("set static text _", Directive {
                    name: "set static text _".into(),
                    doc: "Set the current page's text.\nNormally not used directly; script speech lines call this directive implicitly.".into(),
                    prefix: "set static text".into(),
                    prefix_param: Some(Param {
                        name: "text".into(),
                        doc: Some("Line content.".into()),
                        typ: Type::Speech,
                        nuance: None,
                    }),
                    ..Default::default()
                }),
                ("set static speaker _", Directive {
                    name: "set static speaker _".into(),
                    doc: "Set the current page's speaker.\nNormally not used directly; script speech lines call this directive implicitly.".into(),
                    prefix: "set static speaker".into(),
                    prefix_param: Some(Param {
                        name: "name".into(),
                        doc: Some("Speaker name.".into()),
                        typ: Type::Text,
                        nuance: None,
                    }),
                    ..Default::default()
                }),
                ("characters", Directive {
                    name: "characters".into(),
                    doc: "Set the names of characters in the current scene.".into(),
                    prefix: "characters".into(),
                    list_param: Some(Param {
                        name: "characters".into(),
                        doc: Some("Name of the character to include.".into()),
                        typ: Type::Text,
                        nuance: Some(Box::new(Param {
                            name: "variant".into(),
                            doc: Some("Variant of this character.".into()),
                            typ: Type::Text,
                            nuance: None,
                        })),
                    }),
                    ..Default::default()
                }),
                ("scene", Directive {
                    name: "scene".into(),
                    doc: "Mark the beginning of a scene.".into(),
                    prefix: "scene".into(),
                    ..Default::default()
                }),
                ("background", Directive {
                    name: "background".into(),
                    doc: "Set the background image.".into(),
                    prefix: "background".into(),
                    prefix_param: Some(Param {
                        name: "asset".into(),
                        doc: Some("Background image asset.".into()),
                        typ: Type::Asset { tags: vec_into!["background"] },
                        nuance: None,
                    }),
                    ..Default::default()
                }),
                ("enter _ stage _", Directive {
                    name: "enter _ stage _".into(),
                    doc: "Bring a character onto the stage at a given position.".into(),
                    prefix: "enter".into(),
                    prefix_param: Some(Param {
                        name: "character".into(),
                        doc: Some("Character to enter.".into()),
                        typ: Type::Character,
                        nuance: None,
                    }),
                    suffix: Some("stage".into()),
                    suffix_param: Some(Param {
                        name: "position".into(),
                        doc: Some("Position on the stage where the character enters.".into()),
                        typ: Type::Enum(vec_into!["close", "left third", "right third", "left", "center", "right", "far left", "mid left", "mid right", "far right", "far"]),
                        nuance: None,
                    }),
                    ..Default::default()
                }),
                ("enter _", Directive {
                    name: "enter _".into(),
                    doc: "Bring a character onto the stage at a default position.".into(),
                    prefix: "enter".into(),
                    prefix_param: Some(Param {
                        name: "character".into(),
                        doc: Some("Character to enter.".into()),
                        typ: Type::Character,
                        nuance: None,
                    }),
                    ..Default::default()
                }),
                ("exit _", Directive {
                    name: "exit _".into(),
                    doc: "Remove a character from the stage.".into(),
                    prefix: "exit".into(),
                    prefix_param: Some(Param {
                        name: "character".into(),
                        doc: Some("Character to exit.".into()),
                        typ: Type::Character,
                        nuance: None,
                    }),
                    ..Default::default()
                }),
                ("exeunt", Directive {
                    name: "exeunt".into(),
                    doc: "Remove all characters from the stage.".into(),
                    prefix: "exeunt".into(),
                    ..Default::default()
                }),
            ].map(|(k, v)| (k.into(), v))),
        };
        assert_eq!(s, expected);
    }
}
